#lang racket/base

(require megaparsack megaparsack/text)
(require data/monad)
(require data/applicative)
(require racket/list)
(require racket/function)
(require racket/bool)

(define input-file "input")

(define string-to-eof/p
  (do [ret <- (many/p letter/p)]
    eof/p
    (pure ret)))

(define puzzle-line/p
  (do [min <- integer/p]
    (char/p #\-)
    [max <- integer/p]
    space/p
    [character <- letter/p]
    (char/p #\:)
    space/p
    [password <- string-to-eof/p]
    (pure (hash "min" min "max" max "char" character "pass" password))))

(define (count-char input char)
  (cond
    [(empty? input) 0]
    [(equal? char (first input)) (+ 1 (count-char (rest input) char))]
    [else (count-char (rest input) char)]))


(define (password-line-valid? is-valid? line)
  (let ([parsed (parse-result! (parse-string puzzle-line/p line))])
    (is-valid? parsed)))

(define (get-valid-passwords-from-file is-valid? file)
  (with-input-from-file file
    (thunk
     (for/list ([line (in-lines)]
                #:when (password-line-valid? is-valid? line))
       line))))

(define (count-valid-passwords-from-file is-valid? file)
  (length (get-valid-passwords-from-file is-valid? file)))

(define (correct-num-chars? parsed)
  (let ([min (hash-ref parsed "min")]
        [max (hash-ref parsed "max")]
        [num (count-char (hash-ref parsed "pass") (hash-ref parsed "char"))])
    (and (>= num min)
         (<= num max))))

(define (correct-char-pos? parsed)
  (let ([pos1 (- (hash-ref parsed "min") 1)]
        [pos2 (- (hash-ref parsed "max") 1)]
        [pass (hash-ref parsed "pass")]
        [char (hash-ref parsed "char")])
    (xor (equal? char (list-ref pass pos1))
        (equal? char (list-ref pass pos2)))))

'(cons (count-valid-passwords-from-file correct-num-chars? input-file)
      (count-valid-passwords-from-file correct-char-pos? input-file))

(count-valid-passwords-from-file correct-char-pos? input-file)
